unit Events;
interface
uses Windows;

type
  TCriticalSection = TRTLCriticalSection;
  TEventProcedure = Procedure;

procedure SetMouseEventHandler(p : TEventProcedure);
procedure SetKeyboardEventHandler(p : TEventProcedure);
procedure SetFocusEventHandler(p : TEventProcedure);
procedure SetMenuEventHandler(p : TEventProcedure);
procedure SetResizeEventHandler(p : TEventProcedure);
procedure SetUnknownEventHandler(p : TEventProcedure);

{ these procedures must be used to get the event handlers }
function GetMouseEventHandler : TEventProcedure;
function GetKeyboardEventHandler : TEventProcedure;
function GetFocusEventHandler : TEventProcedure;
function GetMenuEventHandler : TEventProcedure;
function GetResizeEventHandler : TEventProcedure;
function GetUnknownEventHandler : TEventProcedure;

implementation
uses Dos, Ext_Crt;

const
   { these procedures are called if an event occurs }
   MouseEventHandler : procedure = nil;
   KeyboardEventHandler : procedure = nil;
   FocusEventHandler : procedure = nil;
   MenuEventHandler : procedure = nil;
   ResizeEventHandler : procedure = nil;
   UnknownEventHandler  : procedure = nil;

   { if this counter is zero, the event handler thread is killed }
   InstalledHandlers : Byte = 0;

var
   HandlerChanging : TCriticalSection;
   OldExitProc : Pointer;
{$IFDEF VER120}
   EventThreadHandle : THandle;
{$ELSE}
   EventThreadHandle : Handle;
{$ENDIF}
   EventThreadID : DWord;

   { true, if the event handler should be stoped }
   ExitEventHandleThread : boolean;

function GetMouseEventHandler : TEventProcedure;
begin
  GetMouseEventHandler:=MouseEventHandler;
end;


function GetKeyboardEventHandler : TEventProcedure;
begin
  GetKeyboardEventHandler:=KeyboardEventHandler;
end;


function GetFocusEventHandler : TEventProcedure;
begin
  GetFocusEventHandler:=FocusEventHandler;
end;


function GetMenuEventHandler : TEventProcedure;
begin
  GetMenuEventHandler:=MenuEventHandler;
end;


function GetResizeEventHandler : TEventProcedure;
begin
  GetResizeEventHandler:=ResizeEventHandler;
end;


function GetUnknownEventHandler : TEventProcedure;
begin
  GetUnknownEventHandler:=UnknownEventHandler;
end;

{ removes an event from the event queue }
{ necessary, if no handler is installed }
procedure DestroyOneEvent;
var
  ir : TInputRecord;
  dwRead : DWord;
begin
  ReadConsoleInput(hStdin,ir,1,dwRead);
end;

function EventHandleThread(p : pointer) : DWord;StdCall;
var
   ir : TInputRecord;
   dwRead : DWord;
begin
  while not(ExitEventHandleThread) do
  begin
    { wait for an event }
    WaitForSingleObject(hStdin,INFINITE);
    { guard this code, else it is doomed to crash, if the
          thread is switched between the assigned test and
          the call and the handler is removed}
    if not(ExitEventHandleThread) then
    begin
      EnterCriticalSection(HandlerChanging);
      { read, but don't remove the event }
      if PeekConsoleInput(hStdin,ir,1,dwRead) and (dwRead>0) then { call the handler }
      case ir.EventType of
                  KEY_EVENT:
                    begin
                       if assigned(KeyboardEventHandler) then
                         KeyboardEventHandler
                       else
                         DestroyOneEvent;
                    end;
                  _MOUSE_EVENT:
                    begin
                       if assigned(MouseEventHandler) then
                         MouseEventHandler
                       else
                         DestroyOneEvent;
                    end;
                  WINDOW_BUFFER_SIZE_EVENT:
                    begin
                       if assigned(ResizeEventHandler) then
                         ResizeEventHandler
                       else
                         DestroyOneEvent;
                    end;
                  MENU_EVENT:
                    begin
                       if assigned(MenuEventHandler) then
                         MenuEventHandler
                       else
                         DestroyOneEvent;
                    end;
                  FOCUS_EVENT:
                    begin
                       if assigned(FocusEventHandler) then
                         FocusEventHandler
                       else
                         DestroyOneEvent;
                    end;
                  else
                    begin
                       if assigned(UnknownEventHandler) then
                         UnknownEventHandler
                       else
                         DestroyOneEvent;
                    end;
               end;
             LeaveCriticalSection(HandlerChanging);
          end;
     end;
end;

Procedure NewEventHandlerInstalled(p,oldp : TEventProcedure);
var
  oldcount : Byte;
  ir : TInputRecord;
  written : DWord;
begin
  oldcount:=InstalledHandlers;
  if Assigned(oldp) then
    dec(InstalledHandlers);
  if Assigned(p) then
    inc(InstalledHandlers);
  { start event handler thread }
  if (oldcount=0) and (InstalledHandlers=1) then
    begin
       ExitEventHandleThread:=false;
       EventThreadHandle:=CreateThread(nil,0,@EventHandleThread,
         nil,0,EventThreadID);
    end
  { stop and destroy event handler thread }
  else if (oldcount=1) and (InstalledHandlers=0) then
    begin
       ExitEventHandleThread:=true;
       { create a dummy event and sent it to the thread, so
         we can leave WatiForSingleObject }
       ir.EventType:=_MOUSE_EVENT;
       WriteConsoleInput(hStdin,ir,1,written);
       { wait, til the thread is ready }
       WaitForSingleObject(EventThreadHandle,INFINITE);
       CloseHandle(EventThreadHandle);
    end;
end;


Procedure SetMouseEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=MouseEventHandler;
  MouseEventHandler:=p;
  NewEventHandlerInstalled(MouseEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure SetKeyboardEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=KeyboardEventHandler;
  KeyboardEventHandler:=p;
  NewEventHandlerInstalled(KeyboardEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure SetFocusEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=FocusEventHandler;
  FocusEventHandler:=p;
  NewEventHandlerInstalled(FocusEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure SetMenuEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=MenuEventHandler;
  MenuEventHandler:=p;
  NewEventHandlerInstalled(MenuEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure SetResizeEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=ResizeEventHandler;
  ResizeEventHandler:=p;
  NewEventHandlerInstalled(ResizeEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure SetUnknownEventHandler(p : TEventProcedure);
var
  oldp : TEventProcedure;
begin
  EnterCriticalSection(HandlerChanging);
  oldp:=UnknownEventHandler;
  UnknownEventHandler:=p;
  NewEventHandlerInstalled(UnknownEventHandler,oldp);
  LeaveCriticalSection(HandlerChanging);
end;


Procedure DoExit;
begin
  { Uninstall all handlers                   }
  { this stops also the event handler thread }
  SetMouseEventHandler(nil);
  SetKeyboardEventHandler(nil);
  SetFocusEventHandler(nil);
  SetMenuEventHandler(nil);
  SetResizeEventHandler(nil);
  SetUnknownEventHandler(nil);
  { delete the critical section object }
  DeleteCriticalSection(HandlerChanging);
  ExitProc:=OldExitProc;
end;

begin
   InitializeCriticalSection(HandlerChanging);
   OldExitProc:=ExitProc;
   ExitProc:=@DoExit;
end.

